/* cmdline.c
   Program for general command line interface
   Requires Expat (commonly packaged as xmltok-dev)
   Copyright (C) 2003 Linus Walleij

   This file is part of the shftool package.

   Shftool is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   You should have received a copy of the GNU General Public License
   along with Shftool; see the file COPYING.  If not, write to
   the Free Software Foundation, 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA. 

*/

#include "common.h"

/* Extra includes used by this file only */
#include <ctype.h>
#include <stdarg.h>
#include <dirent.h>
#include <sys/stat.h>

// Inputs and outputs
static int input_type = SHF_INPUT;
static int output_type = HEX_OUTPUT;

// input/output file
static FILE* InFile;
static FILE* OutFile;


/**
 * Private function to print usage information.
 */
static void usage() {
  printf("Usage: shftool [-options] <file[.shf]>\n\n");
  printf("Options: -                write to stdout\n");
  printf("         -v               be verbose\n");
  // Print any tool-specific help text
  SHF_print_help();
  HEX_print_help();
  SREC_print_help();
}

void OutString(char *fmt, ...)
{
  va_list ap;
 
  va_start(ap, fmt);
  vfprintf(OutFile, fmt, ap);
  va_end(ap);
}

void OutError(char *fmt, ...)
{
  va_list ap;
 
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
}

void OutChar(char c, bool err)
{
  if (!err) {
    fputc(c, OutFile);
  } else {
    fputc(c, stderr);
  }
}

/************************************************************************
 * Switching station
 * selects writing function depending on selected output filter
 ************************************************************************/

void write_header(char *name, u_int64_t blocks) {
  switch (output_type) {
  case SHF_OUTPUT:
    SHF_write_header(name, blocks);
    break;
  case HEX_OUTPUT:
    HEX_write_header(name, blocks);
    break;
  case SREC_OUTPUT:
    SREC_write_header(name, blocks);
    break;
  }
}

void write_block(shf_block_t *block) {
  switch (output_type) {
  case SHF_OUTPUT:
    SHF_write_block(block);
    break;
  case HEX_OUTPUT:
    HEX_write_block(block);
    break;
  case SREC_OUTPUT:
    SREC_write_block(block);
    break;
  }
}

void write_footer() {
  switch (output_type) {
  case SHF_OUTPUT:
    SHF_write_footer();
    break;
  case HEX_OUTPUT:
    HEX_write_footer();
    break;
  case SREC_OUTPUT:
    SREC_write_footer();
    break;

  }
}

/************************************************************************
 * Main function call to translators
 ************************************************************************/

int
main(int argc, char **argv) {
  int arg, opt;
  
  // Signifies if the program should write to stdout instead of a file
  bool Stdout = false;
  // Dump to output
  bool DumpMode = false;
  char cInFile[STR_MAXLEN] = "";
  char cOutFile[STR_MAXLEN] = "";
  static char cProgramName[]= PACKAGE "-" VERSION;
  extern int optind;
  extern char *optarg;

  while ( (opt = getopt(argc, argv, "i:o:D:vh?" SHF_OPTIONS HEX_OPTIONS SREC_OPTIONS)) != -1 ) {
    switch (opt) {
    case 'D':
      debuglevel = atoi(optarg);
      break;
    case 'v':
      // Verbosity is just the basic debug options.
      debuglevel = 1;
      break;
    case 'i':
      if (!strcmp(optarg,"shf")) {
	input_type = SHF_INPUT;
      } else if (!strcmp(optarg,"srec")) {
	input_type = SREC_INPUT;
      }
      break;
    case 'o':
      if (!strcmp(optarg,"shf")) {
	output_type = SHF_OUTPUT;
      } else if (!strcmp(optarg,"hex")) {
	output_type = HEX_OUTPUT;
      } else if (!strcmp(optarg,"srec")) {
	output_type = SREC_OUTPUT;
      }
      break;
    case 'h':
    case '?':
      usage();
      exit(0);
    default:
      SHF_parse_options(opt, optarg);
      HEX_parse_options(opt, optarg);
      SREC_parse_options(opt, optarg);
      break;
    }
  }
  argc -= optind;
  argv += optind;

  // Check...
  if (argc == 0) {
    usage();
    exit(0);
  }

  // Detect stdout redirector
  if (!strcmp(argv[0], "-")) {
    Stdout = true;
    argc --;
    argv ++;
  }

  // Check again...
  if (argc == 0) {
    usage();
    exit(0);
  }

  // See that the source is OK
  strcpy(cInFile, argv[0]);

  if(!Stdout) {
    strcpy(cOutFile, cInFile);
    // Chop off the filename extension
    if (strlen(cOutFile) > 6) {
      if (cOutFile[strlen(cOutFile)-4] == '.') {
	cOutFile[strlen(cOutFile)-4] = '\0';
      } else if (cOutFile[strlen(cOutFile)-5] == '.') {
	cOutFile[strlen(cOutFile)-5] = '\0';
      }
    }
    // Force new apropriate filename extension
    switch (output_type) {
    case SHF_OUTPUT:
      strncat(cOutFile, ".shf", STR_MAXLEN);
      break;
    case HEX_OUTPUT:
      strncat(cOutFile, ".hex", STR_MAXLEN);
      break;
    case SREC_OUTPUT:
      strncat(cOutFile, ".srec", STR_MAXLEN);
      break;
    default:
      break;
    }
    OutError("Input file: %s\n", cInFile);
    OutError("Output file: %s\n", cOutFile);
    if (!strcmp(cInFile, cOutFile)) {
      OutError("Input and output file are the same!\n");
      OutError("Aborting.\n");
      exit(1);
    }
  }

  // Scan for the file or just assume it is here and open it 
  InFile = fopen(cInFile, "r");
  if (InFile == NULL) {
    OutError("Could not open input file <%s!>\n", cInFile);
    exit(-1);
  }

  if(!Stdout) {
    OutFile = fopen(cOutFile, "w");
    if (OutFile == NULL) {
      OutError("Could not create output file <%s>!\n", cOutFile);
      exit(-1);
    }
  } else {
    OutFile = stdout;
  }

  // Set up
  execution_address_valid = 0;

  // Process the input file with the corresponding parser
  switch(input_type) {
  case SHF_INPUT:
    SHF_process_file(InFile);
    break;
  case SREC_INPUT:
    SREC_process_file(InFile);
    break;
  default:
    OutError("Unknown input type.\n");
    break;
  }

  fclose(InFile);
  if (!Stdout)
    fclose(OutFile);
  exit(0);
}
