/* srec.c
   S-record parser, reader and writer for the SHF tool.
   Copyright (C) 2005 Linus Walleij

   This file is part of the shftool package.

   Shftool is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   You should have received a copy of the GNU General Public License
   along with Shftool; see the file COPYING.  If not, write to
   the Free Software Foundation, 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA. 

*/

#include "common.h"

// Funny datatypes
typedef struct srec_struct srec_t;
struct srec_struct {
  u_int32_t address;
  u_int8_t length;
  u_int8_t *data;
  u_int8_t checksum;
  srec_t *next;
  srec_t *prev;
};

// The width of the addresses: 2,3 or 4 bytes.
static int address_width = 4;

// This list holds the scanned S-records.
static srec_t *srecord_list;
static srec_t *current_entry;

// The name of the dump, if present.
static char dumpname[21];

/**
 * Get the S-record-like checksum for a byte array.
 * Observe that the length here implicitly includes
 * the checksum byte itself, that is why we iterate
 * through to length - 1.
 */
static u_int8_t get_srec_checksum(u_int8_t *array, u_int8_t length)
{
  u_int8_t checksum;
  u_int8_t i;

  checksum = length;
  for (i = 0; i < length-1; i++) {
    checksum += array[i];
  }
  // One's complement the checksum
  checksum ^= 0xff;

  return checksum;
}

/**
 * This function destroys an S-record struct
 */
static void destroy_srec_t(srec_t *record)
{
  if (record == NULL) {
    return;
  }
  if (record->data != NULL) {
    free(record->data);
  }
  free(record);
}

/**
 * This function destroys an entire list of S-record
 * structs.
 */
static void destroy_srec_list(srec_t *first)
{
  srec_t *currentrec;

  if (first == NULL) {
    return;
  }
  currentrec = first;
  while (currentrec != NULL) {
    srec_t *nextrec = currentrec->next;

    destroy_srec_t(currentrec);
    currentrec = nextrec;
  }
}

/**
 * This adds a record to the S-record list
 */
void add_to_recordlist(srec_t *record)
{
  if (current_entry == NULL && srecord_list == NULL) {
    srecord_list = record;
    current_entry = record;
  } else {
    record->prev = current_entry;
    current_entry->next = record;
    current_entry = record;
  }
}

/**
 * This reads a single row from the file, up to any EOL marker.
 */
static int readline_srecfile(FILE *infile, u_int8_t *buffer, int bufsize)
{
  u_int8_t currbyte = '\0';
  int readbytes = 0;

  // File problems - return immediately
  if (feof(infile) || ferror(infile)) {
    return 0;
  }

  // Read first byte off the stream.
  fread(&currbyte, 1, 1, infile);

  // Wind past any newlines.
  while (currbyte == '\n' || currbyte == '\r' && !feof(infile) && !ferror(infile)) {
    fread(&currbyte, 1, 1, infile);
  }

  // Read bytes until a newline or EOF occurs.
  while (currbyte != '\n' && currbyte != '\r' && !feof(infile) && !ferror(infile)) {
    buffer[readbytes] = currbyte;
    readbytes ++;
    if (readbytes == bufsize) {
      OutError("Buffer overflow in S-record reader!\n");
      return -1;
    }
    fread(&currbyte, 1, 1, infile);
  }
  return readbytes;
}

/**
 * This function reads a single S record from the infile
 * and adds it to the linked list.
 */
static void read_s_record(FILE *infile)
{
  u_int8_t srecbuf[515];
  int readbytes = 128;
  u_int8_t payload[256];
  u_int8_t payloadlen;

  // Scan past any non-S-record lines
  srecbuf[0] = '\0';
  while ((readbytes > 0) && (srecbuf[0] != 'S')) {
    readbytes = readline_srecfile(infile, &srecbuf[0], sizeof(srecbuf));
    // Make sure it is properly terminated.
    srecbuf[sizeof(srecbuf)-1] = '\0';
  }
  // Analyze the record, if it was read in successfully.
  if (srecbuf[0] == 'S') {
    if (readbytes < 4) {
      OutError("Ignored invalid S-record (less than 4 characters wide)\n");
    } else {
      u_int8_t numb[3];
      u_int8_t checksum;
      
      // Get the number of bytes in this record.
      numb[0] = srecbuf[2];
      numb[1] = srecbuf[3];
      numb[2] = '\0';
      if ((filterHex(numb[0]) == '\0') || (filterHex(numb[0]) == '\0')) {
	OutError("S-record with non-hex length indicator (\"%s\"). Aborting.\n", (char *) numb);
      } else {
	int i;
	
	payloadlen = hexToNybble(numb[0]) << 4 | hexToNybble(numb[1]);
	if (debuglevel > 0) {
	  OutError("Found S-record with %d bytes.\n", payloadlen);
	}
	for (i = 0; i < payloadlen; i++) {
	  if ((filterHex(srecbuf[4 + i*2]) == '\0') || (filterHex(srecbuf[4 + i*2 +1]) == '\0')) {
	    OutError("S-record with non-hex payload (\"%s\"). Aborting.\n", (char *) srecbuf);
	  } else {
	    payload[i] = hexToNybble(srecbuf[4 + i*2]) << 4 | hexToNybble(srecbuf[4 + i*2 + 1]);
	  }
	}
	// Checksum the payload..
	checksum = get_srec_checksum(payload, payloadlen);
	if (checksum != payload[payloadlen-1]) {
	  OutError("Bad checksum for S-record: Aborting.\n");
	  OutError("Was: %02x, Expected %02x\n", checksum, payload[payloadlen-1]);
	  OutError("%s\n", srecbuf);
	  exit(1);
	}
      }
      
      switch(srecbuf[1]) {
      case '0':
	// This will be used as name for the dump.
	if (payloadlen < 4) {
	  OutError("Bogus header (too few bytes), ignoring S-record:\n");
	  OutError("%s\n", srecbuf);
	} else {
	  // Only use the first 20 characters.
	  if (payloadlen > 23) {
	    payloadlen = 23;
	  }
	  memcpy(dumpname, &payload[2], payloadlen-3);
	  dumpname[payloadlen-3] = '\0';
	}
	break;
      case '1':
	{
	  srec_t *record = (srec_t *) malloc(sizeof(srec_t));
	
	  record->address = payload[0] << 8 | payload[1];
	  // Payload size minus two bytes address and one byte checksum.
	  record->length = payloadlen - 3;
	  record->data = (u_int8_t *) malloc(payloadlen - 3);
	  memcpy(record->data, &payload[2], payloadlen - 3);
	  record->checksum = payload[payloadlen-1];
	  record->next = NULL;
	  record->prev = NULL;
	  add_to_recordlist(record);
	}
	break;
      case '2':
	{
	  srec_t *record = (srec_t *) malloc(sizeof(srec_t));
	
	  record->address = payload[0] << 16 | payload[1] << 8 | payload[2];
	  // Payload size minus threee bytes address and one byte checksum.
	  record->length = payloadlen - 4;
	  record->data = (u_int8_t *) malloc(payloadlen - 4);
	  memcpy(record->data, &payload[3], payloadlen - 4);
	  record->checksum = payload[payloadlen-1];
	  record->next = NULL;
	  record->prev = NULL;
	  add_to_recordlist(record);
	}
	break;
      case '3':
	{
	  srec_t *record = (srec_t *) malloc(sizeof(srec_t));
	  
	  record->address = payload[0] << 24 | payload[1] << 16 | payload[2] << 8 | payload[3];
	  // Payload size minus four bytes address and one byte checksum.
	  record->length = payloadlen - 5;
	  record->data = (u_int8_t *) malloc(payloadlen - 5);
	  memcpy(record->data, &payload[4], payloadlen - 5);
	  record->checksum = payload[payloadlen-1];
	  record->next = NULL;
	  record->prev = NULL;
	  add_to_recordlist(record);
	}
	break;
      case '5':
	/* Count of the number of records transmitted. Just ignore this. */
	break;
      case '7':
	/* Start address with 4 bytes */
	if (payloadlen != 5) {
	  OutError("Bogus start address (too few or too many bytes), ignoring S-record:\n");
	  OutError("%s\n", srecbuf);
	} else {
	  execution_address = (u_int64_t) payload[0] << 24 | payload[1] << 16 | payload[2] << 8 | payload[3];
	  execution_address_valid = 1;
	}
	break;
      case '8':
	/* Start address with 3 bytes */
	if (payloadlen != 4) {
	  OutError("Bogus start address (too few or too many bytes), ignoring S-record:\n");
	  OutError("%s\n", srecbuf);
	} else {
	  execution_address = (u_int64_t) payload[0] << 16 | payload[1] << 8 | payload[2];
	  execution_address_valid = 1;
	}
	break;
      case '9':
	/* Start address with 2 bytes */
	if (payloadlen != 3) {
	  OutError("Bogus start address (too few or too many bytes), ignoring S-record:\n");
	  OutError("%s\n", srecbuf);
	} else {
	  execution_address = (u_int64_t) payload[0] << 8 | payload[1];
	  execution_address_valid = 1;
	}
	break;
      default:
	OutError("Discarding unknown S-record:\n%s\n", (char *) srecbuf);
	break;
      }
    }
  }
  return;
}

/**
 * This takes the list of S records, catenates several S records
 * it into single blocks where possible, creates a corresponding
 * list of SHFtool internal blocks, and exports them.
 */
static void export_shf_blocks(srec_t *list)
{
  srec_t *listp;
  shf_block_t *block_list;
  shf_block_t *current_block;
  u_int64_t blockcount = 0;

  listp = list;
  block_list = NULL;
  current_block = NULL;

  while (listp != NULL) {
    srec_t *first;
    srec_t *current;
    srec_t *last;
    u_int64_t blocksize;
    u_int64_t i;
    shf_block_t *block;

    first = listp;
    current = first;
    // Scan for catenatable region
    while (current->next != NULL &&
	   current->next->address == current->address + current->length) {
      current = current->next;
    }
    last = current;
    
    // Catenate from first to last and form a block
    blocksize = 0;
    current = first;
    while (current != last) {
      blocksize += current->length;
      current = current->next;
    }
    // Add length of last block
    blocksize += current->length;
    // Create SHF block
    block = (shf_block_t *) malloc(sizeof(shf_block_t));
    block->number = blockcount;
    block->name = strdup("S-record block");
    block->address = (u_int64_t) first->address;
    block->word_length = 1;
    block->length = blocksize;
    block->data = (u_int8_t *) malloc(blocksize);
    block->datap = 0;
    // Copy bytes
    current = first;
    while (current != last) {
      memcpy(&block->data[block->datap], current->data, current->length);
      block->datap += current->length;
      current = current->next;
    }
    memcpy(&block->data[block->datap], current->data, current->length);
    block->datap = 0;

    // Create SHA-1 checksum for entire block.
    initialize_sha1();
    for (i = 0; i < block->length; i++) {
      add_byte_to_sha1(block->data[i]);
    }
    block->checksum = retrieve_sha1(block->length);
    blockcount ++;

    // Add to list
    block->next = NULL;
    if (block_list == NULL && current_block == NULL) {
      block_list = block;
      current_block = block;
    } else {
      current_block->next = block;
      current_block = block;
    }
    
    // Advance.
    if (last->next == NULL) {
      listp = NULL;
    } else {
      listp = last->next;
    }
  }

  // Now dump out the blocks and destroy them afterwards.
  if (strlen(dumpname) > 0) {
    write_header(dumpname, blockcount);
  } else {
    write_header("S-record dump", blockcount);
  }
  current_block = block_list;
  while (current_block != NULL) {
    shf_block_t *tmp;

    write_block(current_block);
    tmp = current_block;
    current_block = current_block->next;
    destroy_block_t(tmp);
  }
  write_footer();

  return;
}

/****************************************************************
 * This delimiter marks the beginning of exported "non-static"
 * functions.
 ****************************************************************/

/**
 * Print any tool specific help 
 */
void SREC_print_help(void)
{
  printf("         -i srec          input S-records\n");
  printf("         -o srec          output S-records\n");
  printf("         -s <num>         address width for output S-records: 2, 3 or 4 bytes\n");
}

/** 
 * Parse S-record specific options 
 */
void SREC_parse_options(int opt, char *optarg)
{
  /* Default to something */
  address_width = 4;
  switch (opt) {
  case 's':
    address_width = atoi(optarg);
    if (address_width != 2 && address_width != 3 && address_width != 4) {
      OutError("Illegal address width for S-records, must be 2, 3 or 4.\n");
      OutError("Will use default value of 4.\n");
      address_width = 4;
    }
    if (debuglevel > 0) {
      OutError("Address width of S-records set to %d.\n", address_width);
    }
    break;
  default:
    break;
  }
  return;
}

/**
 * Write a S-record header
 */
void SREC_write_header(char *name, u_int64_t blocks)
{
  char tmpname[25];
  int namelen;
  u_int8_t checksum;
  int i;

  // Adress 0x0000 (needed for checksum)
  tmpname[0] = 0x00U;
  tmpname[1] = 0x00U;
  /* Only use first 20 characters of dump if more. */
  strncpy(tmpname+2, name, 20);
  tmpname[24] = '\0';
  namelen = strlen(tmpname+2);
  OutString("S0%02x0000", namelen+3);
  dumpHexArray((u_int8_t *) tmpname+2, namelen, false);
  // Checksum the dump name
  checksum = get_srec_checksum((u_int8_t *) tmpname, namelen+3);
  OutString("%02x\n", checksum);
  return;
}

/**
 * Write a SHF block as S-records, taking into account
 * the current address width.
 */
void SREC_write_block(shf_block_t *block)
{
  u_int64_t remain = block->length;
  u_int64_t written = 0;
  u_int8_t recordmax;
  
  // If blocksize is too large to fit in current address width,
  // then yell and die or increase it if possible.
  if (remain >> 32 != 0) {
    OutError("The size of block %d of this file is too large to fit into\n", block->number);
    OutError("the largest S-records (S3, 32bit addressing).\n");
    OutError("Aborting.\n");
    exit(1);
  } else if ((remain >> 24 != 0) && address_width < 4) {
    OutError("The size of block %d exceeds the address width of the current\n", block->number);
    OutError("selected S-record type. Increasing to maximum (32bit) width.\n");
    address_width = 4;
  } else if ((remain >> 16 != 0) && address_width < 3) {
    // Do not increase to 3 bytes here, it's awkward and rarely seen. User has to
    // specify that as a command line option if s/he wants it.
    OutError("The size of block %d exceeds the address width of the current\n", block->number);
    OutError("selected S-record type. Increasing to maximum (32bit) width.\n");
    address_width = 4;
  }

  // The maximum number of bytes consumed for each record.
  // Currently hardwired to 0x10 (16) bytes per record.
  // TODO: make this optional? Or use some heuristic, like:
  //       recordmax = 0x13 - address_width - 1;
  recordmax = 0x10;

  while (remain > 0) {
    u_int8_t record[0x20];
    u_int8_t recordlen;
    u_int8_t checksum;

    if (remain < recordmax) {
      recordlen = remain;
    } else {
      recordlen = recordmax;
    }

    // Always use bigendian here.
    // Create an option switch to use little-endian?
    if (address_width == 2) {
      record[0] = ((block->address + written) >> 8) & 0xffU;
      record[1] = (block->address + written) & 0xffU;
    } else if (address_width == 3) {
      record[0] = ((block->address + written) >> 16) & 0xffU;
      record[1] = ((block->address + written) >> 8) & 0xffU;
      record[2] = (block->address + written) & 0xffU;
    } else if (address_width == 4) {
      record[0] = ((block->address + written) >> 24) & 0xffU;
      record[1] = ((block->address + written) >> 16) & 0xffU;
      record[2] = ((block->address + written) >> 8) & 0xffU;
      record[3] = (block->address + written) & 0xffU;
    }

    memcpy(&record[address_width], &block->data[written], recordlen);

    // Split into minor blocks with address width as selected.
    if (address_width == 2) {
      OutString("S1%02x", recordlen + address_width + 1);
    } else if (address_width == 3) {
      OutString("S2%02x", recordlen + address_width + 1);
    } else if (address_width == 4) {
      OutString("S3%02x", recordlen + address_width + 1);
    }

    dumpHexArray(record, recordlen + address_width, false);
    // Checksum the record
    checksum = get_srec_checksum(record, recordlen + address_width + 1);
    OutString("%02x\n", checksum);

    remain -= recordlen;
    written += recordlen;
  }
}

/**
 * Write the footer of a S-record file.
 */
void SREC_write_footer()
{
  // TODO: Optionally write out a S5 summary record?

  // If there is a execution start address, set it here.
  if (execution_address_valid != 0) {
    u_int8_t tmpadd[4];
    u_int8_t checksum;

    if (address_width == 2) {
      tmpadd[0] = (execution_address >> 8) & 0xffU;
      tmpadd[1] = execution_address & 0xffU;
    } else if (address_width == 3) {
      tmpadd[0] = (execution_address >> 16) & 0xffU;
      tmpadd[1] = (execution_address >> 8) & 0xffU;
      tmpadd[2] = execution_address & 0xffU;
    } else if (address_width == 4) {
      tmpadd[0] = (execution_address >> 24) & 0xffU;
      tmpadd[1] = (execution_address >> 16) & 0xffU;
      tmpadd[2] = (execution_address >> 8) & 0xffU;
      tmpadd[3] = execution_address & 0xffU;
    }

    if (address_width == 2) {
      OutString("S903");
    } else if (address_width == 3) {
      OutString("S804");
    } else if (address_width == 4) {
      OutString("S705");
    }

    dumpHexArray(tmpadd, address_width, false);
    // Checksum the address
    checksum = get_srec_checksum(tmpadd, address_width + 1);
    OutString("%02x\n", checksum);    
  }
}

/***
 * Main function for reading S-record files
 */
void SREC_process_file(FILE *infile)
{
  // Reset local record pointers
  srecord_list = NULL;
  current_entry = NULL;
  dumpname[0] = '\0';
  execution_address_valid = 0;
  while (!(feof(infile) || ferror(infile))) {
    read_s_record(infile);
  }
  export_shf_blocks(srecord_list);
  destroy_srec_list(srecord_list);
  return;
}

/* End exported functions */
